/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package org.apache.directory.shared.client.api.operations;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.naming.ldap.StartTlsRequest;

import org.apache.directory.ldap.client.api.LdapConnection;
import org.apache.directory.ldap.client.api.future.ExtendedFuture;
import org.apache.directory.ldap.client.api.message.ExtendedRequest;
import org.apache.directory.ldap.client.api.message.ExtendedResponse;
import org.apache.directory.server.annotations.CreateLdapServer;
import org.apache.directory.server.annotations.CreateTransport;
import org.apache.directory.server.core.CoreSession;
import org.apache.directory.server.core.integ.AbstractLdapTestUnit;
import org.apache.directory.server.core.integ.FrameworkRunner;
import org.apache.directory.server.ldap.handlers.extended.StartTlsHandler;
import org.apache.directory.shared.ldap.message.ResultCodeEnum;
import org.apache.directory.shared.ldap.name.DN;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;

/**
 * Tests the extended operation
 *
 * @author <a href="mailto:dev@directory.apache.org">Apache Directory Project</a>
 * @version $Rev$, $Date$
 */
@RunWith(FrameworkRunner.class)
@CreateLdapServer (
   transports =
   {
       @CreateTransport( protocol = "LDAP" ),
       @CreateTransport( protocol = "LDAPS" )
   },
   extendedOpHandlers =
   {
       StartTlsHandler.class
   })
public class ClientExtendedRequestTest extends AbstractLdapTestUnit
{
   private LdapConnection connection;
   private CoreSession session;

   @Before
   public void setup() throws Exception
   {
       connection = new LdapConnection( "localhost", ldapServer.getPort() );
       DN bindDn = new DN( "uid=admin,ou=system" );
       connection.bind( bindDn.getName(), "secret" );

       session = ldapServer.getDirectoryService().getSession();
   }


   /**
    * Close the LdapConnection
    */
   @After
   public void shutdown()
   {
       try
       {
           if ( connection != null )
           {
               connection.close();
           }
       }
       catch( Exception ioe )
       {
           fail();
       }
   }


   @Test
   public void testExtended() throws Exception
   {
       ExtendedResponse response = connection.extended( StartTlsRequest.OID );
       assertNotNull( response );
       assertEquals( ResultCodeEnum.SUCCESS, response.getLdapResult().getResultCode() );
   }


   @Test
   public void testExtendedAsync() throws Exception
   {
       ExtendedRequest extendedRequest = new ExtendedRequest( StartTlsRequest.OID );

       ExtendedFuture extendedFuture = connection.extendedAsync( extendedRequest );

       try
       {
           ExtendedResponse extendedResponse = extendedFuture.get( 1000, TimeUnit.MILLISECONDS );

           assertNotNull( extendedResponse );
           assertEquals( ResultCodeEnum.SUCCESS, extendedResponse.getLdapResult().getResultCode() );
           assertTrue( connection.isAuthenticated() );
       }
       catch ( TimeoutException toe )
       {
           fail();
       }
   }
}